/*
 *	Copyright (c) 2013 Douglas Miller
 *	$Id: w7ldfix.c,v 1.5 2014-01-01 18:00:21 drmiller Exp $
 *
 * Do post-ld processing (fixup) of Wang 700 Register addresses.
 * Direct references must be semi-BCD, plus add ability to enter
 * indirect register addresses by symbol name.
 *
 * The undefined (illegal) op-code sequence 14-xx RR-RR is used to prefix (escape)
 * a register address, where 'xx' is the operation (+, -, *, /, ST, RE, EXCHG, ST_Y, RE_Y)
 * and RR-RR is the 8-bit binary register address (easily generated by ld).
 * The escape is converted into the appropriate code 04-xx or 12-xx based
 * on range of register (0-99, 100+), and then the adjusted BCD register number.
 * A special case 14-13 14-13 RR-RR is used to identify the "ENTER_REGNO"
 * sequence, where all three steps will be converted to the 3-digit sequence
 * required to enter the indirect register number into X.
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

#include "wang700opcodes.h"

static void w7reginterlace(uint8_t *buf, int len) {
	extern void w7interlacepair(uint8_t *out, uint8_t *reg1, uint8_t *reg2);
	static uint8_t postamble[] = { _op_MARK, _op_STOP, _op_END_PROG, _op_END_PROG };
	static uint8_t fill[] = { _op_STOP, _op_STOP, _op_STOP, _op_STOP,
				_op_STOP, _op_STOP, _op_STOP, _op_STOP };
	if (len <= sizeof(postamble) + 2) {
		// can't recognize image...
		// (actually must be much larger than this, too)
		return;
	}
	uint8_t *post = buf + len - sizeof(postamble);

	if (memcmp(post, postamble, sizeof(postamble)) != 0) {
		// can't recognize image...
		return;
	}
	post -= 2;
	// post[0] is label proctector...
	uint8_t total_space = post[1];
	uint8_t *regs = post - total_space * 8;
	while (regs < post) {
		int padding = (memcmp(&regs[8], fill, 8) == 0);
		w7interlacepair(&regs[0], &regs[0], padding ? NULL : &regs[8]);
		regs += 16;
	}
}

static void w7ldfixup(uint8_t *buf, int len) {
	uint8_t *end = buf + len;
	while (buf < end) {
		len = end - buf;
		if (len >= 3 && buf[0] == _tag_last_reg && buf[1] == _tag_last_reg) {
			int reg = buf[2];
			buf[0] = _pre_E | (reg / 100);
			buf[1] = _pre_E | ((reg / 10) % 10);
			buf[2] = _pre_E | (reg % 10);
			buf += 3;
		} else if (len >= 2 && (buf[0] & _mask_DIR) == _tag_reg_dir ) {
			int reg = buf[1];
			if (reg >= 100) {
				buf[0] = _pre_DIR100 | (buf[0] & _mask_DIROP);
				reg -= 100;
			} else {
				buf[0] = _pre_DIR | (buf[0] & _mask_DIROP);
			}
			buf[1] = ((reg / 10) << 4) | (reg % 10);
			buf += 2;
		} else if (len >= 2 && (buf[0] & ~_mask_2STEP) == _id_2STEP) {
			if (buf[0] == _op_END_ALPHA ||
					(buf[0] > _op_EXCHG_DIR100 &&
					buf[0] < _op_ST_Y_DIR100)) {
				buf += 1;
			} else {
				// need to track WRITE_ALPHA .. END_ALPHA?
				buf += 2;
			}
		} else {
			buf += 1;
		}
	}
}

int main(int argc, char **argv) {
	int x;
	char *t;
	if (argc < 2) {
		fprintf(stderr, "Usage: %s <file> [...]\n", argv[0]);
		exit(1);
	}
	for (x = 1; x < argc; ++x) {
		t = strrchr(argv[x], '.');
		if (t == NULL || strcmp(t, ".w7$") != 0) {
			fprintf(stderr, "%s: filename must match *.w7$\n", argv[x]);
			exit(1);
		}
		int fd = open(argv[x], O_RDONLY);
		if (fd < 0) {
			perror(argv[x]);
			exit(1);
		}
		struct stat stb;
		fstat(fd, &stb);
		uint8_t *buf = malloc(stb.st_size);
		if (buf == NULL) {
			perror("malloc");
			exit(1);
		}
		int n = read(fd, buf, stb.st_size);
		if (n != stb.st_size) {
			if (n < 0) perror(argv[x]);
			else fprintf(stderr, "%s: incomplete read\n", argv[x]);
			exit(1);
		}
		close(fd);
		w7ldfixup(buf, stb.st_size);
		w7reginterlace(buf, stb.st_size);
		strcpy(t, ".w7t");
		fd = open(argv[x], O_WRONLY | O_CREAT | O_TRUNC, 0666);
		if (fd < 0) {
			perror(argv[x]);
			exit(1);
		}
		n = write(fd, buf, stb.st_size);
		if (n != stb.st_size) {
			if (n < 0) perror(argv[x]);
			else fprintf(stderr, "%s: incomplete write\n", argv[x]);
			exit(1);
		}
		free(buf);
		n = close(fd);
		if (n < 0) {
			perror(argv[x]);
			exit(1);
		}
	}
	return 0;
}
